// Copyright 2017 Google Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package com.google.webauthn.gaedemo.servlets;

import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.URL;
import java.security.KeyPair;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import co.nstant.in.cbor.CborDecoder;
import co.nstant.in.cbor.CborException;
import co.nstant.in.cbor.model.*;
import com.google.appengine.api.users.UserService;
import com.google.appengine.api.users.UserServiceFactory;
import com.google.common.base.Splitter;
import com.google.common.collect.Iterables;
import com.google.common.io.BaseEncoding;
import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonParseException;
import com.google.gson.JsonParser;
import com.google.webauthn.gaedemo.exceptions.ResponseException;
import com.google.webauthn.gaedemo.objects.AttestationExtension;
import com.google.webauthn.gaedemo.objects.AuthenticatorAttestationResponse;
import com.google.webauthn.gaedemo.objects.CablePairingData;
import com.google.webauthn.gaedemo.objects.CableRegistrationData;
import com.google.webauthn.gaedemo.objects.PublicKeyCredential;
import com.google.webauthn.gaedemo.server.AndroidSafetyNetServer;
import com.google.webauthn.gaedemo.server.PackedServer;
import com.google.webauthn.gaedemo.server.PublicKeyCredentialResponse;
import com.google.webauthn.gaedemo.server.U2fServer;
import com.google.webauthn.gaedemo.service.UserProxyService;
import com.google.webauthn.gaedemo.storage.CableKeyPair;
import com.google.webauthn.gaedemo.storage.Credential;
import kong.unirest.HttpResponse;
import kong.unirest.Unirest;
import org.bouncycastle.util.encoders.HexEncoder;
import unirest.shaded.org.apache.commons.codec.binary.Hex;

public class FinishMakeCredential extends HttpServlet {

	private static final int FINGERPRINT = 2;
	private static final int SCREEN_LOCK = 134;
	private static final long serialVersionUID = 1L;
	private final UserService userService = UserServiceFactory.getUserService();

	public FinishMakeCredential() {}

	@Override
	protected void doGet(HttpServletRequest request, HttpServletResponse response)
			throws ServletException, IOException {
		doPost(request, response);
	}

	@Override
	protected void doPost(HttpServletRequest request, HttpServletResponse response)
			throws ServletException, IOException {
		/* Tyler start -----------------*/
		UserProxyService userProxyService = UserProxyService.getInstance();
		String currentUser = userProxyService.getCurrentUser().getEmail();
		/* Tyler end -----------------*/
    /* Original code
    String currentUser = userService.getCurrentUser().getEmail();
     */
		String data = request.getParameter("data");
		if (data == null) {
			data = "";
		}
//		System.out.println(data);
		String session = request.getParameter("session");
		if (session == null) {
			session = "";
		}

		String credentialId = null;
		String type = null;
		String uvm = null;
		JsonElement makeCredentialResponse = null;
		CablePairingData cablePairingData = null;

		try {
			JsonObject json = new JsonParser().parse(data).getAsJsonObject();
			JsonElement idJson = json.get("id");
			if (idJson != null) {
				credentialId = idJson.getAsString();
			}
			JsonElement typeJson = json.get("type");
			if (typeJson != null) {
				type = typeJson.getAsString();
			}
			JsonElement uvmJson = json.get("uvm");
			if (uvmJson != null && uvmJson.isJsonArray()) {
				JsonArray uvmArray = uvmJson.getAsJsonArray();
				if (uvmJson.isJsonArray()) {
					JsonElement uvmElement = uvmArray.get(0);
					if (uvmElement != null) {
						switch (uvmElement.getAsJsonObject().get("userVerificationMethod").getAsInt()){
							case FINGERPRINT:
								uvm = "Fingerprint";
								break;
							case SCREEN_LOCK:
								uvm = "Screen Lock";
								break;
							default:
								uvm = "Others";
								break;
						}
					}
				}
			}
			makeCredentialResponse = json.get("response");
		} catch (IllegalStateException e) {
			throw new ServletException("Passed data not a json object");
		} catch (ClassCastException e) {
			throw new ServletException("Invalid input");
		} catch (JsonParseException e) {
			throw new ServletException("Input not valid json");
		}

		AuthenticatorAttestationResponse attestation = null;
		try {
			attestation = new AuthenticatorAttestationResponse(makeCredentialResponse);
		} catch (ResponseException e) {
			throw new ServletException(e);
		}

		if (attestation.getAttestationObject().getAuthenticatorData().hasExtensionData()) {
			Map<String, AttestationExtension> extensionMap =
					attestation.getAttestationObject().getAuthenticatorData().getExtensionData();
			if (extensionMap.containsKey(CableRegistrationData.KEY)) {
				CableRegistrationData cableData =
						(CableRegistrationData) extensionMap.get(CableRegistrationData.KEY);

				// Get key pair generated during the StartMakeCredential operation
				KeyPair sessionKeyPair = CableKeyPair.get(Long.valueOf(session));

				cablePairingData = CablePairingData.generatePairingData(cableData, sessionKeyPair);
			}
		}

		/* josh start */
		String IDfromIDP = "";
		if(1==0) {
			String attestationObjectBase64 = makeCredentialResponse.getAsJsonObject().get("attestationObject").getAsString();
			System.out.println("attestationObject base64: " + attestationObjectBase64);
			byte[] attestationObject = BaseEncoding.base64().decode(attestationObjectBase64);
			try {
				co.nstant.in.cbor.model.Map attestationObjectCbor = (co.nstant.in.cbor.model.Map) CborDecoder.decode(attestationObject).get(0);
				for (DataItem key : attestationObjectCbor.getKeys())
					System.out.println(key);

				co.nstant.in.cbor.model.Map attestationStatement = (co.nstant.in.cbor.model.Map) attestationObjectCbor.get(new UnicodeString("attStmt"));
				for (DataItem key : attestationStatement.getKeys())
					System.out.println(key);

				List<DataItem> extensions = ((Array) attestationStatement.get(new UnicodeString("extensions"))).getDataItems();
				for (DataItem item : extensions)
					System.out.println(item);

				byte[][] extensionSendsOut = new byte[2][];
				extensionSendsOut[0] = ((ByteString) extensions.get(0)).getBytes();
				extensionSendsOut[1] = ((ByteString) extensions.get(1)).getBytes();

				for (byte[] bytes : extensionSendsOut)
					System.out.println(Hex.encodeHexString(bytes));

				String p1 = BaseEncoding.base64().encode(extensionSendsOut[0]); // param 1 : hmac
				String p2 = BaseEncoding.base64().encode(extensionSendsOut[1]); // param 2 : Cx

				HttpResponse<String> httpResponse = Unirest.post("http://localhost:8086/api/idp/askIdentity")
						.header("Content-Type", "application/json")
						.body("{\"p1\":\"" + p1 + "\",\"p2\":\"" + p2 + "\"}")
						.asString();

				IDfromIDP = httpResponse.getBody();
				System.out.println("the identity returned from IDP : " + IDfromIDP);
			} catch (CborException e) {
				e.printStackTrace();
			}
		}/* josh end */

		// Recoding of credential ID is needed, because the ID from HTTP servlet request doesn't support
		// padding.
		String credentialIdRecoded =
				BaseEncoding.base64Url().encode(BaseEncoding.base64Url().decode(credentialId));

		PublicKeyCredential cred = new PublicKeyCredential(credentialIdRecoded, type,
				BaseEncoding.base64Url().decode(credentialId), attestation);

		String domain = (request.isSecure() ? "https://" : "http://") + request.getHeader("Host");
		String rpId = Iterables.get(Splitter.on(':').split(request.getHeader("Host")), 0);
		switch (cred.getAttestationType()) {
			case FIDOU2F:
				U2fServer.registerCredential(cred, currentUser, session, domain, rpId);
				break;
			case ANDROIDSAFETYNET:
				AndroidSafetyNetServer.registerCredential(cred, currentUser, session, rpId);
				break;
			case PACKED:
//        PackedServer.registerCredential(cred, currentUser, session, rpId); //
				break;
			case NONE:
				break;
		}

		Credential credential = new Credential(cred);
		if (cablePairingData != null) {
			credential.setCablePairingData(cablePairingData);
		}
		credential.setUserVerificationMethod(uvm);
		credential.save(currentUser);

		PublicKeyCredentialResponse rsp =
				new PublicKeyCredentialResponse(true, "Successfully created credential, Hi "+IDfromIDP);

		response.setContentType("application/json");
		response.getWriter().println(rsp.toJson());
	}

}
